
package visitor;

import mainpackage.Main;
import syntaxtree.*;


public class MaxTempVisitor<R, A> extends GJDepthFirst<R, A>
{
    private String currentFunction;
    
    /**
     * f0 -> "MAIN" 
     * f1 -> StmtList() 
     * f2 -> "END" 
     * f3 -> ( Procedure() )* 
     * f4 -> <EOF>
     */
    public R visit(Goal n, A argu)
    {
        R _ret = null;                       
        n.f0.accept(this, argu);
        
        currentFunction = "MAIN";
        
        n.f1.accept(this, argu);
        n.f2.accept(this, argu);
        n.f3.accept(this, argu);
        n.f4.accept(this, argu);                
        
        return _ret;
    }   

    /**
     * f0 -> Label() 
     * f1 -> "[" 
     * f2 -> IntegerLiteral() 
     * f3 -> "]" 
     * f4 -> StmtExp()
     */
    public R visit(Procedure n, A argu)
    {
        R _ret = null;
        n.f0.accept(this, argu);
        
        currentFunction = n.f0.f0.tokenImage;
        
        n.f1.accept(this, argu);
        n.f2.accept(this, argu);
        n.f3.accept(this, argu);
        n.f4.accept(this, argu);
        return _ret;
    }    

    /**
     * f0 -> "TEMP" 
     * f1 -> IntegerLiteral()
     */
    public R visit(Temp n, A argu)
    {
        R _ret = null;
        n.f0.accept(this, argu);
        n.f1.accept(this, argu);
                
        int temp = Integer.parseInt(n.f1.f0.tokenImage);
        if(Main.maxTemps.containsKey(currentFunction))
        {            
            // If the current temp value is greater than last max it becomes max            
            if(Main.maxTemps.get(currentFunction) < temp)
            {
                Main.maxTemps.put(currentFunction, temp);
            }
        }       
        else
        {
            Main.maxTemps.put(currentFunction, temp);
        }
        
        return _ret;
    }
}
