
package visitor;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import mainpackage.Main;
import syntaxtree.*;


public class PigletVisitor extends GJDepthFirst<String, String>
{
    private String currentFunction;
    private List<String> callArgsList = new ArrayList(); 
    
    /**
     * f0 -> "MAIN" 
     * f1 -> StmtList() 
     * f2 -> "END" 
     * f3 -> ( Procedure() )* 
     * f4 -> <EOF>
     */
    public String visit(Goal n, String argu)
    {
        String _ret = null;
        
        try
        {            
            n.f0.accept(this, argu);
            
            currentFunction = "MAIN";
            Main.writer.write("MAIN ");
            
            n.f1.accept(this, argu);                        
            n.f2.accept(this, argu);
            
            Main.writer.write("END ");
            
            n.f3.accept(this, argu);
            n.f4.accept(this, argu);                        
        }
        catch (IOException ex)
        {
            System.out.println(ex.getMessage());
        }
        
        return _ret;
    }

    /**
     * f0 -> ( ( Label() )? Stmt() )*
     */
    public String visit(StmtList n, String argu)
    {
        return n.f0.accept(this, "print label");
    }

    /**
     * f0 -> Label() 
     * f1 -> "[" 
     * f2 -> IntegerLiteral() 
     * f3 -> "]" 
     * f4 -> StmtExp()
     */
    public String visit(Procedure n, String argu)
    {
        String _ret = null;
        String label = n.f0.accept(this, argu);  
        
        currentFunction = n.f0.f0.tokenImage;
        
        n.f1.accept(this, argu);        
        
        String il = n.f2.accept(this, argu);
        
        try
        {
            Main.writer.write(label + " [" + il.trim() + "] ");
        }
        catch (IOException ex)
        {
            System.err.println(ex.getMessage());
        }                
        
        n.f3.accept(this, argu);
        n.f4.accept(this, "print fun");
        
        return _ret;
    }

    /**
     * f0 -> NoOpStmt() 
     *     | ErrorStmt() 
     *     | CJumpStmt() 
     *     | JumpStmt() 
     *     | HStoreStmt()
     *     | HLoadStmt() 
     *     | MoveStmt() 
     *     | PrintStmt()
     */
    public String visit(Stmt n, String argu)
    {
        String stmt = n.f0.accept(this, null);
        try
        {
            Main.writer.write(stmt);
        }
        catch (IOException ex)
        {
            System.err.println(ex.getMessage());
        }
        
        return null;
    }

    /**
     * f0 -> "NOOP"
     */
    public String visit(NoOpStmt n, String argu)
    {          
        n.f0.accept(this, argu);
        
        return n.f0.tokenImage + " ";
    }

    /**
     * f0 -> "ERROR"
     */
    public String visit(ErrorStmt n, String argu)
    {
        n.f0.accept(this, argu);
        
        return n.f0.tokenImage + " ";
    }

    /**
     * f0 -> "CJUMP" 
     * f1 -> Exp() 
     * f2 -> Label()
     */
    public String visit(CJumpStmt n, String argu)
    {
        String _ret;
        
        n.f0.accept(this, argu);

        String exp = n.f1.accept(this, argu);

        // If the expression is a temp 
        if(exp.startsWith("TEMP"))
        {
            _ret = "CJUMP " + exp;
        }
        else
        {
            // Get the max temp plus 1 and make an intermidiate command
            // to pass the result of the expression in a temp
            int t = Main.maxTemps.get(currentFunction) + 1;
            _ret = "MOVE TEMP " + t + " " + exp + "CJUMP TEMP " + t + " ";
            Main.maxTemps.put(currentFunction, t);
        }

        _ret += n.f2.accept(this, argu);           
        
        return _ret;
    }

    /**
     * f0 -> "JUMP" 
     * f1 -> Label()
     */
    public String visit(JumpStmt n, String argu)
    {
        String _ret;        
        
        n.f0.accept(this, argu);                
        
        _ret = "JUMP " + n.f1.accept(this, argu);
        
        return _ret;
    }

    /**
     * f0 -> "HSTORE" 
     * f1 -> Exp() 
     * f2 -> IntegerLiteral() 
     * f3 -> Exp()
     */
    public String visit(HStoreStmt n, String argu)
    {
        String _ret = "";

        n.f0.accept(this, argu);            

        String exp = n.f1.accept(this, argu);
        String exp1 = null, exp2 = null;
        int t1 = 0, t2 = 0;

        // If the expression is a temp
        if(exp.startsWith("TEMP"))
        {
            exp1 = exp;
        }
        else
        {
            // Get the max temp plus 1 and make an intermidiate command
            // to pass the result of the expression in a temp
            t1 = Main.maxTemps.get(currentFunction) + 1;
            _ret += "MOVE TEMP " + t1 + " " + exp;
            Main.maxTemps.put(currentFunction, t1);                
        }            

        String il = n.f2.accept(this, argu);

        exp = n.f3.accept(this, argu);     

        // If the expression is a temp
        if(exp.startsWith("TEMP"))
        {
            exp2 = exp;
        }
        else
        {
            // Get the max temp plus 1 and make an intermidiate command
            // to pass the result of the expression in a temp            
            t2 = Main.maxTemps.get(currentFunction) + 1;
            _ret += "MOVE TEMP " + t2 + " " + exp;
            Main.maxTemps.put(currentFunction, t2);                
        }   
        
        _ret += "HSTORE " + (exp1 == null ? "TEMP " + t1 + " ": exp1)
                + il + (exp2 == null ? "TEMP " + t2 + " ": exp2);
        
        return _ret;
    }

    /**
     * f0 -> "HLOAD" 
     * f1 -> Temp() 
     * f2 -> Exp() 
     * f3 -> IntegerLiteral()
     */
    public String visit(HLoadStmt n, String argu)
    {
        String _ret;     
        
        n.f0.accept(this, argu);
        
        String temp = n.f1.accept(this, argu);
        
        String exp = n.f2.accept(this, argu);
        
        // If the expression is a temp
        if(exp.startsWith("TEMP"))
        {
            _ret = "HLOAD " + temp + exp;
        }
        else
        {
            // Get the max temp plus 1 and make an intermidiate command
            // to pass the result of the expression in a temp            
            int t = Main.maxTemps.get(currentFunction) + 1;
            _ret = "MOVE TEMP " + t + " " + exp + "HLOAD " + temp + "TEMP " + t + " ";
            Main.maxTemps.put(currentFunction, t);                
        }        
        
        _ret += n.f3.accept(this, argu);
        return _ret;
    }

    /**
     * f0 -> "MOVE" 
     * f1 -> Temp() 
     * f2 -> Exp()
     */
    public String visit(MoveStmt n, String argu)
    {
        String _ret = "MOVE ";        
        
        n.f0.accept(this, argu);
        
        _ret += n.f1.accept(this, argu);
        _ret += n.f2.accept(this, argu);
        
        return _ret;
    }

    /**
     * f0 -> "PRINT" 
     * f1 -> Exp()
     */
    public String visit(PrintStmt n, String argu)
    {
        String _ret;       
        
        n.f0.accept(this, argu);
        
        String exp = n.f1.accept(this, argu);
        
        // If the exprression is a SimpleExp
        if(isSimpleExp(exp))
        {
            _ret = "PRINT " + exp;             
        }
        else
        {
            // Get the max temp plus 1 and make an intermidiate command
            // to pass the result of the expression in a temp            
            int t = Main.maxTemps.get(currentFunction) + 1;
            _ret = "MOVE TEMP " + t + " " + exp + "PRINT " + "TEMP " + t + " ";
            Main.maxTemps.put(currentFunction, t);
        }
        
        return _ret;
    }

    /**
     * f0 -> StmtExp() 
     *     | Call() 
     *     | HAllocate() 
     *     | BinOp() 
     *     | Temp() 
     *     | IntegerLiteral() 
     *     | Label()
     */
    public String visit(Exp n, String argu)
    {
        String exp = n.f0.accept(this, null);
                
        // If we are in a CALL instruction
        if(argu != null)
        {
            int pos = Integer.parseInt(argu);            
            // If the expression is a temp 
            if(exp.startsWith("TEMP"))
            {
                //put it on the arguments list
                callArgsList.set(pos, callArgsList.get(pos) + exp);
            }
            else
            {
                // Get the max temp plus 1 and make an intermidiate command
                // to pass the result of the expression in a temp                
                int t = Main.maxTemps.get(currentFunction) + 1;
                try
                {
                    Main.writer.write("MOVE TEMP " + t + " " + exp);
                }
                catch (IOException ex)
                {
                    System.err.println(ex.getMessage());
                }                
                // then put it on the arguments list
                callArgsList.set(pos, callArgsList.get(pos) + "TEMP " + t + " ");
                Main.maxTemps.put(currentFunction, t);                
            }             
        }
        
        return exp;
    }

    /**
     * f0 -> "BEGIN" 
     * f1 -> StmtList() 
     * f2 -> "RETURN" 
     * f3 -> Exp() 
     * f4 -> "END"
     */
    public String visit(StmtExp n, String argu)
    {
        String _ret = null;
        
        // If we are at a StmtExp of a function
        if(argu != null && argu.equals("print fun"))
        {
            try
            {
                Main.writer.write("BEGIN ");
            }
            catch (IOException ex)
            {
                System.err.println(ex.getMessage());
            }
        }
        
        n.f0.accept(this, null);
        n.f1.accept(this, null);                       
        n.f2.accept(this, null);
        
        
        String exp = n.f3.accept(this, null);
        // If we are at a StmtExp of a function
        if(argu != null && argu.equals("print fun"))
        {
            // If the expression is a SimpleExp
            if(isSimpleExp(exp))
            {
                exp = "RETURN " + exp + "END ";             
            }
            else
            {
                // Get the max temp plus 1 and make an intermidiate command
                // to pass the result of the expression in a temp                
                int t = Main.maxTemps.get(currentFunction) + 1;
                exp = "MOVE TEMP " + t + " " + exp + "RETURN " + "TEMP " + t + " END ";
                Main.maxTemps.put(currentFunction, t);
            }            
            
            try
            {
                Main.writer.write(exp);
            }
            catch (IOException ex)
            {
                System.err.println(ex.getMessage());
            }
        }    
        else
        {
            // If the expression is a SimpleExp
            if(isSimpleExp(exp))
            {
                _ret = exp;             
            }
            else
            {
                // Get the max temp plus 1 and make an intermidiate command
                // to pass the result of the expression in a temp                
                int t = Main.maxTemps.get(currentFunction) + 1;
                exp = "MOVE TEMP " + t + " " + exp;
                Main.maxTemps.put(currentFunction, t);
                
                try
                {
                    Main.writer.write(exp);
                }
                catch (IOException ex)
                {
                    System.err.println(ex.getMessage());
                }                
                
                _ret = "TEMP " + t;
            }                                   
        }
        
        n.f4.accept(this, null);
        
        return _ret;
    }

    /**
     * f0 -> "CALL" 
     * f1 -> Exp() 
     * f2 -> "(" 
     * f3 -> ( Exp() )* 
     * f4 -> ")"
     */
    public String visit(Call n, String argu)
    {
        String _ret;
        
        n.f0.accept(this, argu);
        
        String exp = n.f1.accept(this, argu);
        
        // If the expression is a SimplExp
        if(isSimpleExp(exp))
        {
            _ret = "CALL " + exp.trim() + "( ";             
        }
        else
        {
            // Get the max temp plus 1 and make an intermidiate command
            // to pass the result of the expression in a temp            
            int t = Main.maxTemps.get(currentFunction) + 1;
            try
            {
                Main.writer.write("MOVE TEMP " + t + " " + exp);
            }
            catch (IOException ex)
            {
                System.err.println(ex.getMessage());
            }                        
            Main.maxTemps.put(currentFunction, t);
            
            _ret = "CALL " + "TEMP " + t + "( ";
        } 
        
        n.f2.accept(this, argu);
        
        // Add a new element on the arguments list and pass the position of it
        // to the next function, then retrieve the arguments that were inserted
        int pos = callArgsList.size();
        callArgsList.add(pos, "");
        n.f3.accept(this, "" + pos);        
        _ret += callArgsList.get(pos) + ") ";        
        
        n.f4.accept(this, argu);
        
        return _ret;
    }

    /**
     * f0 -> "HALLOCATE" 
     * f1 -> Exp()
     */
    public String visit(HAllocate n, String argu)
    {
        String _ret = null;
        
        n.f0.accept(this, argu);
        String exp = n.f1.accept(this, argu);
        
        // If the expression is a SimpleExp
        if(isSimpleExp(exp))
        {
            _ret = "HALLOCATE " + exp;
        }
        else
        {
            // Get the max temp plus 1 and make an intermidiate command
            // to pass the result of the expression in a temp            
            int t = Main.maxTemps.get(currentFunction) + 1;
            try
            {
                Main.writer.write("MOVE TEMP " + t + " " + exp);
            }
            catch (IOException ex)
            {
                System.err.println(ex.getMessage());
            }                        
            Main.maxTemps.put(currentFunction, t);                
            
            _ret = "HALLOCATE TEMP " + t + " ";
        }        
        
        return _ret;
    }

    /**
     * f0 -> Operator() 
     * f1 -> Exp() 
     * f2 -> Exp()
     */
    public String visit(BinOp n, String argu)
    {
        String _ret;                
        
        String op = n.f0.accept(this, argu);
        
        String exp = n.f1.accept(this, argu);
        
        String exp1 = null, exp2 = null;
        int t1 = 0, t2 = 0;

        // If the expression is a temp
        if(exp.startsWith("TEMP"))
        {
            exp1 = exp;
        }
        else
        {
            // Get the max temp plus 1 and make an intermidiate command
            // to pass the result of the expression in a temp            
            t1 = Main.maxTemps.get(currentFunction) + 1;
            try
            {
                Main.writer.write("MOVE TEMP " + t1 + " " + exp);
            }
            catch (IOException ex)
            {
                System.err.println(ex.getMessage());
            }
            Main.maxTemps.put(currentFunction, t1);                
        }            

        exp = n.f2.accept(this, argu);

        // If the expression is a SimpleExp
        if(isSimpleExp(exp))
        {
            exp2 = exp;
        }
        else
        {
            // Get the max temp plus 1 and make an intermidiate command
            // to pass the result of the expression in a temp            
            t2 = Main.maxTemps.get(currentFunction) + 1;
            try
            {
                Main.writer.write("MOVE TEMP " + t2 + " " + exp);
            }
            catch (IOException ex)
            {
                System.err.println(ex.getMessage());
            }                        
            Main.maxTemps.put(currentFunction, t2);                
        }   
        
        _ret = op + (exp1 == null ? "TEMP " + t1 + " ": exp1)
                + (exp2 == null ? "TEMP " + t2 + " ": exp2);                        
        
        return _ret;
    }

    /**
     * f0 -> "LT" 
     *     | "PLUS" 
     *     | "MINUS" 
     *     | "TIMES"
     */
    public String visit(Operator n, String argu)
    {
        n.f0.accept(this, argu);
        String op = n.f0.choice.toString() + " ";       
        
        return op;
    }

    /**
     * f0 -> "TEMP" 
     * f1 -> IntegerLiteral()
     */
    public String visit(Temp n, String argu)
    {                    
        n.f0.accept(this, argu);
        String i = n.f1.accept(this, argu) + " ";
        
        return "TEMP " + i;
    }

    /**
     * f0 -> <INTEGER_LITERAL>
     */
    public String visit(IntegerLiteral n, String argu)
    {
        n.f0.accept(this, argu);        
        
        return n.f0.tokenImage + " ";
    }

    /**
     * f0 -> <IDENTIFIER>
     */
    public String visit(Label n, String argu)
    {
        n.f0.accept(this, null); 
        
        // If the we are at the StmtList print the label
        if(argu != null && argu.equals("print label"))
        {
            try
            {
                Main.writer.write(n.f0.tokenImage + " ");
            }
            catch (IOException ex)
            {
                System.out.println(ex.getMessage());
            }             
        }
        
        return n.f0.tokenImage + " ";
    }

    // Function that checks if an expression is a SimpleExp
    private boolean isSimpleExp(String exp)
    {        
        return !exp.startsWith("CALL") && !exp.startsWith("HALLOCATE") && !exp.startsWith("LT")
               && !exp.startsWith("PLUS") && !exp.startsWith("MINUS") && !exp.startsWith("TIMES");
    }
}
